#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <linux/vt.h>
#include "device.h"
#include "fb.h"
/*
 * VT switch handling code for Linux
 */

/* signal to use when VT swithing*/
#define SIGVTSWITCH	SIGUNUSED	/* SIGUSR2 is used by pthreads...*/

int	vterm;			/* the VT we were started on */
int 	cvt, ocvt;
int	drawing;		/* nonzero when drawing is happening*/
static int ttyfd = -1;		/* /dev/tty0*/
static int visible = 1;		/* VT visible flag*/
static struct vt_mode mode;	/* terminal mode*/
static SUBDRIVER save;		/* saved subdriver when VT switched*/

extern SCREENDEVICE	scrdev;	//FIXME

/* entry points*/
int 	fb_InitVt(void);
int  	fb_CurrentVt(void);
int  	fb_CheckVtChange(void);
void 	fb_RedrawVt(int t);

/* local routines*/
static void  	draw_enable(void);
static void 	draw_disable(void);
static void	vt_switch(int sig);

/* null subdriver for drawing when switched out*/
static void 	null_drawpixel(PSD psd,COORD x, COORD y, PIXELVAL c) {}
static PIXELVAL	null_readpixel(PSD psd,COORD x, COORD y) { return 0;}
static void	null_drawhorzline(PSD psd,COORD x1,COORD x2,COORD y,
			PIXELVAL c) {}
static void	null_drawvertline(PSD psd,COORD x,COORD y1,COORD y2,
			PIXELVAL c) {}
static void	null_fillrect(PSD psd,COORD x1,COORD y1,COORD x2,COORD y2,
			PIXELVAL c) {}
static void	null_blit(PSD dstpsd,COORD destx,COORD desty,COORD w,COORD h,
			PSD srcpsd,COORD srcx,COORD srcy,int op) {}
static void 	null_drawarea(PSD psd, driver_gc_t *gc, int op) {}
static SUBDRIVER nulldriver = {
	NULL,
	null_drawpixel,
	null_readpixel,
	null_drawhorzline,
	null_drawvertline,
	null_fillrect,
	null_blit
};

static void
draw_enable(void)
{
	if(visible)
		return;
	visible = 1;

	/* restore screen drawing functions*/
	set_subdriver(&scrdev, &save, FALSE);
}
      
static void
draw_disable(void)
{
	if(!visible)
		return;
	visible = 0;

	/* save screen drawing functions and reroute drawing*/
	get_subdriver(&scrdev, &save);

	/* set null driver*/
	set_subdriver(&scrdev, &nulldriver, FALSE);
}

/* Signal handler called when kernel switches to or from our tty*/
static void
vt_switch(int sig)
{
    struct itimerval duration;
    static unsigned short r[16], g[16], b[16];

    signal(SIGVTSWITCH, vt_switch);

    /* If a drawing function is in progress then we cannot mode
     * switch right now because the drawing function would continue to
     * scribble on the screen after the switch.  So disable further
     * drawing and schedule an alarm to try again in .1 second.
     */
    if(drawing) {
    	draw_disable ();

    	signal(SIGALRM, vt_switch);

	duration.it_interval.tv_sec = 0;
	duration.it_interval.tv_usec = 0;
	duration.it_value.tv_sec = 0;
	duration.it_value.tv_usec = 100000;
    	setitimer (ITIMER_REAL, &duration, NULL);
    	return;
    }
      
    if(visible) {
    	draw_disable ();
	ioctl_getpalette(0, 16, r, g, b);

	if(ioctl (ttyfd, VT_RELDISP, 1) == -1)
	    failmsg("can't switch away from VT: %m");
    } else {
	ioctl_setpalette(0, 16, r, g, b);
    	draw_enable ();
      
	if(ioctl (ttyfd, VT_RELDISP, VT_ACKACQ) == -1)
		failmsg("can't acknowledge VT switch: %m");
    }
}

/*
 * Init VT switch catching code
 * 	return 0 on success, -1 on error
 */
int
fb_InitVt(void)
{
	ttyfd = open("/dev/tty0", O_RDONLY);
	if(ttyfd == -1)
		return failmsg("can't open tty0: %m");
	
	/* setup new tty mode*/
	if(ioctl (ttyfd, VT_GETMODE, &mode) == -1)
		return failmsg("can't get VT mode: %m");

	mode.mode = VT_PROCESS;
	mode.relsig = SIGVTSWITCH;
	mode.acqsig = SIGVTSWITCH;
	signal (SIGVTSWITCH, vt_switch);
	if(ioctl (ttyfd, VT_SETMODE, &mode) == -1)
		return failmsg("can't set VT mode: %m");

	cvt = ocvt = vterm = fb_CurrentVt();
	/*
	 * Note: this hack is required to get Linux
	 * to orient virtual 0,0 with physical 0,0
	 * I have no idea why this kluge is required...
	 */
	fb_RedrawVt(vterm);

	return 0;
}

/*
 * This function is used to find out what the current active VT is.
 */
int
fb_CurrentVt(void)
{
	struct vt_stat stat;

	ioctl(ttyfd, VT_GETSTATE, &stat);
	return stat.v_active;
}

/*
 * Check if our VT has changed.  Return 1 if so.
 */
int
fb_CheckVtChange(void)
{
	cvt = fb_CurrentVt();
	if(cvt != ocvt && cvt == vterm) {
		ocvt = cvt;
		return 1;
	}
	ocvt = cvt;
	return 0;
}

/*
 * This function is used to cause a redraw of the text console.
 * FIXME: Switching to another console and
 * back works, but that's a really dirty hack
 */
void
fb_RedrawVt(int t)
{
	if(fb_CurrentVt() == vterm) {
		ioctl(ttyfd, VT_ACTIVATE, t == 1 ? 2 : 1); /* Awful hack!!*/
		ioctl(ttyfd, VT_ACTIVATE, t);
	}
}

void
fb_ExitVt(void)
{
	signal(SIGVTSWITCH, SIG_DFL);
	mode.mode = VT_AUTO;
	mode.relsig = 0;
	mode.acqsig = 0;
	ioctl(ttyfd, VT_SETMODE, &mode);

	if(ttyfd != -1)
		close(ttyfd);
}
